// This is a generic version of Int_double_pair. Unlike Int_double_pair,
// which holds an int and a double, Pair can hold any two types. In
// particular, Pair<T1, T2> holds a T1 and a T2.

#include <iostream>
#include <algorithm>

template <typename T1, typename T2>
struct Pair
{
    Pair(const T1&, const T2&);

    T1 first;
    T2 second;
};

// Function for constructing a pair from the two values. Unlike using the
// Pair constructor directly, this does not require explicitly specifying the
// types.
template <typename T1, typename T2>
Pair<T1, T2> make_Pair(const T1&, const T2&);

// Stream insertion, for printing pairs.
template <typename T1, typename T2>
std::ostream& operator<<(std::ostream&, const Pair<T1, T2>&);

// Equality and inequality on pairs requires equality on T1 and T2
template <typename T1, typename T2>
bool operator==(const Pair<T1, T2>&, const Pair<T1, T2>&);

template <typename T1, typename T2>
bool operator!=(const Pair<T1, T2>&, const Pair<T1, T2>&);

// Comparisons in pairs require operator< to be defined for both T1 and T2.
template <typename T1, typename T2>
bool operator< (const Pair<T1, T2>&, const Pair<T1, T2>&);

template <typename T1, typename T2>
bool operator<=(const Pair<T1, T2>&, const Pair<T1, T2>&);

template <typename T1, typename T2>
bool operator> (const Pair<T1, T2>&, const Pair<T1, T2>&);

template <typename T1, typename T2>
bool operator>=(const Pair<T1, T2>&, const Pair<T1, T2>&);

template <typename T1, typename T2>
void swap(Pair<T1, T2>&, Pair<T1, T2>&);

/////
///// IMPLEMENTATION
///// When defining templates, the implementation has to go in the header.
/////

template <typename T1, typename T2>
Pair<T1, T2>::Pair(const T1& v1, const T2& v2)
        : first(v1), second(v2)
{ }

template <typename T1, typename T2>
Pair<T1, T2> make_Pair(const T1& v1, const T2& v2)
{
    return {v1, v2};
}

template <typename T1, typename T2>
std::ostream& operator<<(std::ostream& os, const Pair<T1, T2>& p)
{
    return os << '(' << p.first << ", " << p.second << ')';
}

template <typename T1, typename T2>
bool operator==(const Pair<T1, T2>& a, const Pair<T1, T2>& b)
{
    return a.first == b.first && a.second == b.second;
}

template <typename T1, typename T2>
bool operator!=(const Pair<T1, T2>& a, const Pair<T1, T2>& b)
{
    return !(a == b);
}

template <typename T1, typename T2>
bool operator< (const Pair<T1, T2>& a, const Pair<T1, T2>& b)
{
    if (a.first < b.first) return true;
    if (b.first < a.first) return false;
    return a.second < b.second;
}

template <typename T1, typename T2>
bool operator<=(const Pair<T1, T2>& a, const Pair<T1, T2>& b)
{
    return !(b < a);
}

template <typename T1, typename T2>
bool operator> (const Pair<T1, T2>& a, const Pair<T1, T2>& b)
{
    return b < a;
}

template <typename T1, typename T2>
bool operator>=(const Pair<T1, T2>& a, const Pair<T1, T2>& b)
{
    return !(a < b);
}

template <typename T1, typename T2>
void swap(Pair<T1, T2>& p1, Pair<T1, T2>& p2)
{
    using std::swap;

    swap(p1.first, p2.first);
    swap(p1.second, p2.second);
};

/////
///// The get<which>() function
/////

namespace detail {

template <typename T, size_t which>
struct Get;

template <typename T1, typename T2>
struct Get<Pair<T1, T2>, 0>
{
    using pair_type = Pair<T1, T2>;
    using result_type = T1;

    static const result_type& get(const pair_type& p)
    { return p.first; }

    static result_type& get(pair_type& p)
    { return p.first; }
};

template <typename T1, typename T2>
struct Get<Pair<T1, T2>, 1>
{
    using pair_type = Pair<T1, T2>;
    using result_type = T2;

    static const result_type& get(const pair_type& p)
    { return p.second; }

    static result_type& get(pair_type& p)
    { return p.second; }
};

} // end namespace detail

template <size_t which, typename T1, typename T2>
const auto& get(const Pair<T1, T2>& p)
{
    return detail::Get<Pair<T1, T2>, which>::get(p);
};

template <size_t which, typename T1, typename T2>
auto& get(Pair<T1, T2>& p)
{
    return detail::Get<Pair<T1, T2>, which>::get(p);
};

template <typename T1, typename T2>
const T1& get(const Pair<T1, T2>& p)
{
    return p.first;
};

template <typename T1, typename T2>
T1& get(Pair<T1, T2>& p)
{
    return p.first;
};

template <typename T2, typename T1>
const T2& get(const Pair<T1, T2>& p)
{
    return p.second;
};

template <typename T2, typename T1>
T2& get(Pair<T1, T2>& p)
{
    return p.second;
};
